#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <stdexcept>

#include "binary.hpp"

using namespace std;

typedef unsigned long ulong;
typedef unsigned int uint;
typedef unsigned short ushort;
typedef unsigned char uchar;

namespace{
    void fmt_bit(char buf[8], uchar b)
    {
        for(int i = 0; i < 8; ++i)
            buf[i] = (b&(1<<(7-i))) ? '1' : '0';
    }
}

long numdec_s_le(const uchar* buf, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numdec_s_le(): invalid size");
    return static_cast<long>(numdec_u_le(buf, size));
}

ulong numdec_u_le(const uchar* buf, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numdec_u_le(): invalid size");
    ulong result = 0;
    for(int i = 0; i < size; ++i)
        result |= buf[i]<<(8*i);
    return result;
}

long numdec_s_be(const uchar* buf, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numdec_s_be(): invalid size");
    return static_cast<long>(numdec_u_be(buf, size));
}

ulong numdec_u_be(const uchar* buf, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numdec_u_be(): invalid size");
    ulong result = 0;
    for(int i = 0; i < size; ++i)
        result |= buf[i]<<(8*(size-1-i));
    return result;
}

long numdec_s32_le(const uchar* buf) { return numdec_s_le(buf, 4); }
ulong numdec_u32_le(const uchar* buf) { return numdec_u_le(buf, 4); }
long numdec_s24_le(const uchar* buf) { return numdec_s_le(buf, 3); }
ulong numdec_u24_le(const uchar* buf) { return numdec_u_le(buf, 3); }
long numdec_s16_le(const uchar* buf) { return numdec_s_le(buf, 2); }
ulong numdec_u16_le(const uchar* buf) { return numdec_u_le(buf, 2); }
long numdec_s32_be(const uchar* buf) { return numdec_s_be(buf, 4); }
ulong numdec_u32_be(const uchar* buf) { return numdec_u_be(buf, 4); }
long numdec_s24_be(const uchar* buf) { return numdec_s_be(buf, 3); }
ulong numdec_u24_be(const uchar* buf) { return numdec_u_be(buf, 3); }
long numdec_s16_be(const uchar* buf) { return numdec_s_be(buf, 2); }
ulong numdec_u16_be(const uchar* buf) { return numdec_u_be(buf, 2); }

void numenc_s_le(uchar* buf, long n, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numenc_s_le(): invalid size");
    numenc_u_le(buf, static_cast<ulong>(n), size);
}

void numenc_u_le(uchar* buf, ulong n, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numenc_u_le(): invalid size");
    for(int i = 0; i < size; ++i)
        buf[i] = (n>>(8*i))&0xFF;
}

void numenc_s_be(uchar* buf, long n, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numenc_s_be(): invalid size");
    numenc_u_be(buf, static_cast<ulong>(n), size);
}

void numenc_u_be(uchar* buf, ulong n, int size)
{
    if(size < 0 || size > 4) throw invalid_argument("numenc_u_be(): invalid size");
    for(int i = 0; i < size; ++i)
        buf[i] = (n>>(8*(size-1-i)))&0xFF;
}

void numenc_s32_le(uchar* buf, long n) { numenc_s_le(buf, n, 4); }
void numenc_u32_le(uchar* buf, ulong n) { numenc_u_le(buf, n, 4); }
void numenc_s24_le(uchar* buf, long n) { numenc_s_le(buf, n, 3); }
void numenc_u24_le(uchar* buf, ulong n) { numenc_u_le(buf, n, 3); }
void numenc_s16_le(uchar* buf, long n) { numenc_s_le(buf, n, 2); }
void numenc_u16_le(uchar* buf, ulong n) { numenc_u_le(buf, n, 2); }
void numenc_s32_be(uchar* buf, long n) { numenc_s_be(buf, n, 4); }
void numenc_u32_be(uchar* buf, ulong n) { numenc_u_be(buf, n, 4); }
void numenc_s24_be(uchar* buf, long n) { numenc_s_be(buf, n, 3); }
void numenc_u24_be(uchar* buf, ulong n) { numenc_u_be(buf, n, 3); }
void numenc_s16_be(uchar* buf, long n) { numenc_s_be(buf, n, 2); }
void numenc_u16_be(uchar* buf, ulong n) { numenc_u_be(buf, n, 2); }

void strdump_hex(string& dump, const uchar* buf, ulong size)
{
    ostringstream out;

    string delim(" ");
    for(ulong i = 0; i < size; ++i){
        if(i != 0) out << delim;
        out << hex << uppercase << setfill('0') << setw(2);
        out << static_cast<uint>(buf[i]);
    }

    dump = out.str();
}

void strdump_bin(string& dump, const uchar* buf, ulong size)
{
    ostringstream out;

    char bit_buf[8];
    string delim(" ");
    for(ulong i = 0; i < size; ++i){
        if(i != 0) out << delim;
        fmt_bit(bit_buf, buf[i]);
        out.write(bit_buf, 8);
    }

    dump = out.str();
}
