#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include "binary.hpp"

#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const char* FCM_SIGNATURE = "FCM\x1A";
    const uint FCM_HDR_SIZE = 0x34;
    const uint FCM_FLAG_FROMRESET = 0x02;
    const uint FCM_FLAG_PAL = 0x04;
    const uint FCM_FLAG_NOSYNCHACK = 0x10;
    enum FCMCmd{
        FCM_CMD_RESET = 0x01,
        FCM_CMD_POWER = 0x02,
        FCM_CMD_FDSINSERT = 0x18,
        FCM_CMD_FDSSELECT = 0x1A,
    };
    enum FCMButton{
        FCM_BUTTON_A = 0,
        FCM_BUTTON_B = 1,
        FCM_BUTTON_S = 2,
        FCM_BUTTON_T = 3,
        FCM_BUTTON_U = 4,
        FCM_BUTTON_D = 5,
        FCM_BUTTON_L = 6,
        FCM_BUTTON_R = 7,
    };

    // TASVideosŎgĂFPS
    const double FPS_FCEU[2] = { 60.0, 50.0 };
    const double FPS_FCEUX[2] = { 60.098813897440515532, 50.006977968268290849 };

    const char* FCS_SIGNATURE = "FCS";
    const uint FCS_HDR_SIZE = 0x10;
    const uchar FCS_OLDVER_NEWFCEU = 0xFF;

    struct ArgStruct{
        string path_fcm;
        string path_state;
        string path_body;
        bool verbose;
    };

    enum MovieFrom{
        MOVIE_FROM_RESET,
        MOVIE_FROM_POWER,
        MOVIE_FROM_STATE,
    };

    struct FCMInfo{
        MovieFrom from;
        uchar flag;
        ulong frame;
        ulong rerecord;
        ulong stateoff;
        ulong statesize;
        ulong statesize_hdr;
        ulong bodyoff;
        ulong bodysize;
        ulong filesize;
        ulong emu_ver;
        MD5 rom_md5;
        string rom_name;
        string metadata;
        ulong state_ver;
        bool state_ver_new;
        bool port[4];
    };

    void warn(const string& msg)
    {
        cerr << msg << endl;
    }

    void error(const string& msg)
    {
        warn(msg);
        exit(1);
    }

    ulong align4(ulong n)
    {
        return n + ((4-(n&0x03))&0x03);
    }

    string int2hex(ulong n, char fill, int w)
    {
        ostringstream out;
        out << hex << uppercase << setfill(fill) << setw(w) << n;
        return out.str();
    }

    string float2str(double n, int prec)
    {
        ostringstream out;
        out << showpoint << setprecision(prec) << n;
        return out.str();
    }

    void read_chk(istream& in, uchar* buf, streamsize size)
    {
        in.read((char*)buf, size);
        if(!in) error("Incomplete FCM file");
    }

    void read_chk(istream& in, string& buf, streamsize size)
    {
        uchar* tmp = new uchar[size];
        read_chk(in, tmp, size);
        buf.assign((char*)tmp, size);
        delete[] tmp;
    }

    void read_cstr_chk(istream& in, string& buf)
    {
        stringbuf tmp;
        in.get(tmp, '\0');
        char c;
        in.get(c);
        if(c != '\0')
            error("Incomplete FCM file");
        else
            if(in.fail()) in.clear();
        buf = tmp.str();
    }

    void verbose_line(ulong addr, const uchar* buf, uint size, const string& msg)
    {
        // cout ̏ԂύXƊOɉeyԂ̂ŁAostringstream oR
        ostringstream out;

        string dump;
        strdump_hex(dump, buf, size);
        out << "0x" << int2hex(addr, '0', 6) << "  ";
        out << dump << "\t# " << msg;

        cout << out.str() << endl;
    }

    void verbose_line_frame(ulong addr, const uchar* buf, uint size, const string& msg, ulong f)
    {
        ostringstream oss;
        oss << "Frame " << f << ": " << msg;
        verbose_line(addr, buf, size, oss.str());
    }

    void fcm_read(const string& path, FCMInfo& info, string& state, string& body)
    {
        ifstream in(path.c_str(), ios_base::in | ios_base::binary);
        if(!in) error("Can't open input file");
        in.seekg(0, ios_base::end);
        info.filesize = in.tellg();
        in.seekg(0, ios_base::beg);
        if(info.filesize < FCM_HDR_SIZE) error("Incomplete FCM file");

        uchar hdr[FCM_HDR_SIZE];
        read_chk(in, hdr, FCM_HDR_SIZE);
        if(memcmp(hdr+0x00, FCM_SIGNATURE, 4) != 0) error("FCM signature not found");
        ulong fcm_ver = numdec_u32_le(hdr+0x04);
        if(fcm_ver != 2) error("Not FCMv2 file");

        info.flag = hdr[0x08];
        if(info.flag & ~(FCM_FLAG_FROMRESET|FCM_FLAG_PAL|FCM_FLAG_NOSYNCHACK)){
            ostringstream oss;
            string dump;
            strdump_bin(dump, &info.flag, 1);
            oss << "[WARNING] Unknown bit in flag: 0x0008  " << dump;
            warn(oss.str());
        }
        if(hdr[0x09] || hdr[0x0A] || hdr[0x0B]){
            ostringstream oss;
            string dump;
            strdump_hex(dump, hdr+0x09, 3);
            oss << "[WARNING] Unknown bit in reserved flags: 0x0009  " << dump;
            warn(oss.str());
        }

        info.frame = numdec_u32_le(hdr+0x0C);
        info.rerecord = numdec_u32_le(hdr+0x10);

        info.bodysize = numdec_u32_le(hdr+0x14);
        info.stateoff = numdec_u32_le(hdr+0x18);
        info.bodyoff = numdec_u32_le(hdr+0x1C);
        if(info.stateoff >= info.bodyoff) error("Invalid FCM file (stateoff >= bodyoff)");
        if(info.bodyoff + info.bodysize > info.filesize) error("Incomplete FCM file");
        if(info.stateoff != align4(info.stateoff)) warn("[WARNING] State offset not aligned to 4-Byte boundary");
        if(info.bodyoff != align4(info.bodyoff)) warn("[WARNING] Input data offset not aligned to 4-Byte boundary");
        info.statesize = info.bodyoff - info.stateoff;

        md5_from_raw(info.rom_md5, hdr+0x20);
        info.emu_ver = numdec_u32_le(hdr+0x30);

        read_cstr_chk(in, info.rom_name);
        read_cstr_chk(in, info.metadata);
        if(info.rom_name.empty()) warn("[WARNING] Empty ROM name");

        in.seekg(info.stateoff, ios_base::beg);
        read_chk(in, state, info.statesize);

        in.seekg(info.bodyoff, ios_base::beg);
        read_chk(in, body, info.bodysize);

        in.close();
    }

    bool save_file(const string& path, const string& content)
    {
        ofstream out(path.c_str(), ios_base::out | ios_base::trunc | ios_base::binary);
        if(!out) return false;
        out << content;
        out.close();
        return true;
    }

    void chk_state(FCMInfo& info, const string& state)
    {
        if(state.size() < FCS_HDR_SIZE) error("State: incomplete FCS");
        const uchar* p = (const uchar*)state.data();
        if(memcmp(p+0x00, FCS_SIGNATURE, 3) != 0) error("State: FCS signature not found");

        uchar ver_old = *(p+0x03);
        info.statesize_hdr = 16 + numdec_u32_le(p+4);
        if(ver_old == FCS_OLDVER_NEWFCEU){
            info.state_ver_new = true;
            info.state_ver = numdec_u32_le(p+8);
        }
        else{
            info.state_ver_new = false;
            info.state_ver = 100*ver_old;
        }
    }

    const char* fcm_cmd2str(uint cmd)
    {
        switch(cmd){
        case FCM_CMD_RESET: return "Reset";
        case FCM_CMD_POWER: return "Power";
        case FCM_CMD_FDSINSERT: return "FDSInsert";
        case FCM_CMD_FDSSELECT: return "FDSSelect";
        default:
            error("fcm_cmd2str(): invalid argument");
            return NULL;
        }
    }

    const char* fcm_button2str(uint button)
    {
        switch(button){
        case FCM_BUTTON_A: return "A";
        case FCM_BUTTON_B: return "B";
        case FCM_BUTTON_S: return "select";
        case FCM_BUTTON_T: return "start";
        case FCM_BUTTON_U: return "up";
        case FCM_BUTTON_D: return "down";
        case FCM_BUTTON_L: return "left";
        case FCM_BUTTON_R: return "right";
        default:
            error("fcm_button2str(): invalid argument");
            return NULL;
        }
    }

    void chk_body(FCMInfo& info, const string& body, bool verbose)
    {
        istringstream in(body);
        ulong addr = info.bodyoff;
        if(!(info.flag&FCM_FLAG_FROMRESET)){
            info.from = MOVIE_FROM_STATE;
        }
        else{
            uchar first;
            in.get((char&)first);
            if(!in) error("Incomplete FCM input data");
            switch(first){
            case 0x81: info.from = MOVIE_FROM_RESET; break;
            case 0x82: info.from = MOVIE_FROM_POWER; break;
            default: error("No Power/Reset command at beginning of non-state-based movie");
            }
            if(verbose){
                string msg = (info.from==MOVIE_FROM_RESET ? "From reset" : "From power");
                verbose_line(addr, &first, 1, msg);
            }
            ++addr;
        }

        for(int i = 0; i < 4; ++i) info.port[i] = false;

        ulong f = 0;
        uint cmd = 0;
        uchar joy[4] = { 0, 0, 0, 0 }; // RLDUTSBA
        uchar buf[4];
        for(;;){
            in.get((char&)buf[0]);
            if(!in) error("Incomplete FCM input data");

            bool is_cmd = buf[0]&0x80;
            uint deltasize = (buf[0]>>5)&0x03;
            uint data = buf[0]&0x1F;

            ulong delta;
            if(!deltasize){
                delta = 0;
            }
            else{
                read_chk(in, buf+1, deltasize);
                delta = numdec_u_le(buf+1, deltasize);
                cmd = 0;
            }
            f += delta;

            if(is_cmd){
                if(data == 0){
                    if(verbose)
                        verbose_line_frame(addr, buf, 1+deltasize, "Movie end", f);
                    break;
                }

                if(cmd){
                    ostringstream oss;
                    oss << "Frame " << f << ": multiple commands in 1 frame";
                    error(oss.str());
                }

                cmd = data;
                switch(cmd){
                case FCM_CMD_RESET:
                case FCM_CMD_POWER:
                case FCM_CMD_FDSINSERT:
                case FCM_CMD_FDSSELECT:
                    if(verbose){
                        ostringstream oss;
                        oss << "<CMD> " << fcm_cmd2str(cmd);
                        verbose_line_frame(addr, buf, 1+deltasize, oss.str(), f);
                    }
                    break;
                default:
                    ostringstream oss;
                    oss << "Frame " << f << ": unknown command";
                    error(oss.str());
                }
            }
            else{
                uint id = (data>>3)&0x03;
                uint button = data&0x07;

                info.port[id] = true;
                joy[id] ^= 1<<button;
                bool press = joy[id]&(1<<button);

                if(verbose){
                    ostringstream oss;
                    oss << "<P" << 1+id << "> " << (press?"press":"release") << " " << fcm_button2str(button);
                    verbose_line_frame(addr, buf, 1+deltasize, oss.str(), f);
                }
            }

            addr += 1+deltasize;
        }

        if(info.frame != f){
            ostringstream oss;
            oss << "Frame count mismatch (header: " << info.frame << ", actual: " << f << ")";
            error(oss.str());
        }

        if(verbose) cout << endl;
    }

    void frame2time(ulong frame, double fps, string& time)
    {
        ulong csec = (ulong)floor(100.0*frame/fps + 0.5);

        ostringstream out;
        out << setfill('0') << setw(2) << csec/100/(60*60);
        out << ":";
        out << setfill('0') << setw(2) << (csec/100/60)%60;
        out << ":";
        out << setfill('0') << setw(2) << (csec/100)%60;
        out << ".";
        out << setfill('0') << setw(2) << csec%100;
        time = out.str();
    }

    const char* bool2str(bool b)
    {
        return (b?"true":"false");
    }

    const char* from2str(MovieFrom from)
    {
        switch(from){
        case MOVIE_FROM_RESET: return "Reset";
        case MOVIE_FROM_POWER: return "Power";
        case MOVIE_FROM_STATE: return "State";
        default:
            error("from2str(): invalid argument");
            return NULL;
        }
    }

    void report(const FCMInfo& info)
    {
        bool pal = info.flag&FCM_FLAG_PAL;
        bool synchack = !(info.flag&FCM_FLAG_NOSYNCHACK);
        double fps_fceu = FPS_FCEU[(pal?1:0)];
        double fps_fceux = FPS_FCEUX[(pal?1:0)];
        string time_fceu;
        string time_fceux;
        frame2time(info.frame, fps_fceu, time_fceu);
        frame2time(info.frame, fps_fceux, time_fceux);
        string rom_md5_str;
        md5_to_str(info.rom_md5, rom_md5_str);
        ulong garbagesize = info.filesize - (info.bodyoff+info.bodysize);

        cout << "Emulator: FCEU " << info.emu_ver << endl;
        cout << "ROM name: " << info.rom_name << endl;
        cout << "ROM MD5: " << rom_md5_str << endl;
        cout << "Recorded from: " << from2str(info.from) << endl;
        cout << "Timing: " << (pal?"PAL":"NTSC") << endl;
        cout << "Sync hack: " << bool2str(synchack) << endl;
        cout << "Frames: " << info.frame << endl;
        cout << "Time (" << float2str(fps_fceu, 5) << " FPS): " << time_fceu << endl;
        cout << "Time (" << float2str(fps_fceux, 5) << " FPS): " << time_fceux << endl;
        cout << "Rerecords: " << info.rerecord << endl;
        cout << "Author: " << info.metadata << endl;
        cout << "Controller: ";
        for(int i = 0; i < 4; ++i){
            if(i != 0) cout << ", ";
            cout << "P" << 1+i << "=" << bool2str(info.port[i]);
        }
        cout << endl;
        cout << "FCS version: " << info.state_ver << " (" << (info.state_ver_new?"new":"old") << ")" << endl;
        cout << "FCS: "
             << "offset=0x" << int2hex(info.stateoff, '0', 6)
             << ", size(header)=0x" << int2hex(info.statesize_hdr, '0', 6)
             << ", size(actual?)=0x" << int2hex(info.statesize, '0', 6) << endl;
        cout << "Input data: "
             << "offset=0x" << int2hex(info.bodyoff, '0', 6)
             << ", size=0x" << int2hex(info.bodysize, '0', 6) << endl;
        cout << "Garbage at end: " << garbagesize << " Bytes" << endl;
    }

    void usage()
    {
        error("Usage: fcmcheck [-v] [-os FCS] [-ob body] <FCM>");
    }

    void parse_args(int argc, char** argv, ArgStruct& arg)
    {
        arg.verbose = false;

        --argc; ++argv;
        while(argc){
            char* str = *argv;
            if(str[0] != '-') break;

            if(strcmp(str+1, "v") == 0){
                arg.verbose = true;
            }
            else if(strcmp(str+1, "os") == 0){
                --argc; ++argv;
                if(!argc) usage();
                arg.path_state = *argv;
            }
            else if(strcmp(str+1, "ob") == 0){
                --argc; ++argv;
                if(!argc) usage();
                arg.path_body = *argv;
            }
            else{
                ostringstream oss;
                oss << "Invalid option: " << str;
                error(oss.str());
            }
            --argc; ++argv;
        }

        if(argc != 1) usage();
        arg.path_fcm = *argv;
    }
}

int main(int argc, char** argv)
{
    ArgStruct arg;
    parse_args(argc, argv, arg);

    FCMInfo info;
    string state;
    string body;
    fcm_read(arg.path_fcm, info, state, body);

    // ǂݍ݂ɐƂ肠Xe[gƓ̓f[^o(IvVw肪ꍇ)
    // eɌ肪邩ȂAo͂邱Ǝ̂ɖ͂Ȃ̂ŁB
    if(!arg.path_state.empty())
        if(!save_file(arg.path_state, state)) warn("[WARNING] Can't write FCS");
    if(!arg.path_body.empty())
        if(!save_file(arg.path_body, body)) warn("[WARNING] Can't write body");

    chk_state(info, state);
    chk_body(info, body, arg.verbose);

    report(info);

    return 0;
}
