#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdlib>
#include <cstring>

#include <zlib.h>

#include "binary.hpp"

#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const char* FCS_SIGNATURE = "FCS";
    const uint FCS_HDR_SIZE = 0x10;
    const uchar FCS_OLDVER_NEWFCEU = 0xFF;
    const uchar FCS_OLDVER_FCEUX = 0x58; // 'X'
    const ulong FCS_COMPSIZE_UNCOMP = 0xFFFFFFFF;

    struct FCSInfo{
        uchar oldver;
        ulong newver;
        ulong size_hdr;
        bool comp;
        ulong compsize;
        ulong filesize;
    };

    void warn(const string& msg)
    {
        cerr << msg << endl;
    }

    void error(const string& msg)
    {
        warn(msg);
        exit(1);
    }

    bool is_bzero(const uchar* buf, ulong size)
    {
        for(ulong i = 0; i < size; ++i)
            if(buf[i]) return false;
        return true;
    }

    void read_chk(istream& in, uchar* buf, streamsize size)
    {
        in.read((char*)buf, size);
        if(!in) error("Incomplete FCS file");
    }

    void read_chk(istream& in, string& buf, streamsize size)
    {
        uchar* tmp = new uchar[size];
        read_chk(in, tmp, size);
        buf.assign((char*)tmp, size);
        delete[] tmp;
    }

    void fcs_read(const string& path, FCSInfo& info, string& body)
    {
        ifstream in(path.c_str(), ios_base::in | ios_base::binary);
        if(!in) error("Can't open input file");
        in.seekg(0, ios_base::end);
        info.filesize = in.tellg();
        in.seekg(0, ios_base::beg);
        if(info.filesize < FCS_HDR_SIZE) error("Incomplete FCS file");

        uchar hdr[FCS_HDR_SIZE];
        read_chk(in, hdr, FCS_HDR_SIZE);
        if(memcmp(hdr+0x00, FCS_SIGNATURE, 3) != 0) error("FCS signature not found");
        info.oldver = hdr[0x03];
        info.size_hdr = numdec_u32_le(hdr+0x04);

        /**
         * FCEU͋ZNVFCS̃f[^TCY𐳂o͂Ȃ̂ŁA
         * TCYsłG[ƂȂ
         */
        if(info.oldver == FCS_OLDVER_NEWFCEU){
            info.newver = numdec_u32_le(hdr+0x08);
            info.comp = false;
            if(!is_bzero(hdr+0x0C, 4)){
                ostringstream oss;
                string dump;
                strdump_hex(dump, hdr+0x0C, 4);
                oss << "[WARNING] Nonzero in reserved area: 0x000C  " << dump;
                warn(oss.str());
            }
        }
        else if(info.oldver == FCS_OLDVER_FCEUX){
            info.newver = numdec_u32_le(hdr+0x08);
            info.compsize = numdec_u32_le(hdr+0x0C);
            info.comp = !(info.compsize == FCS_COMPSIZE_UNCOMP);
            /**
             * FCEUX͏ɐf[^TCYނ͂Ȃ̂ŁATCYs
             * ̓G[ƂĈ
             */
            ulong expsize = (info.comp ? info.compsize : info.size_hdr);
            if(expsize+FCS_HDR_SIZE != info.filesize){
                ostringstream oss;
                oss << "Filesize mismatch "
                    << "(header: " << expsize+FCS_HDR_SIZE
                    << ", actual: " << info.filesize << ")";
                error(oss.str());
            }
        }
        else{
            info.comp = false;
            if(!is_bzero(hdr+0x08, 8)){
                ostringstream oss;
                string dump;
                strdump_hex(dump, hdr+0x08, 8);
                oss << "[WARNING] Nonzero in reserved area: 0x0008  " << dump;
                warn(oss.str());
            }
        }

        read_chk(in, body, info.filesize-FCS_HDR_SIZE);

        in.close();
    }

    void chk_comp(const FCSInfo& info, const string& body)
    {
        if(!info.comp) return;

        ulong uncompsize = info.size_hdr;
        uchar* buf = new uchar[uncompsize];
        if(Z_OK != uncompress(buf, &uncompsize, (const Bytef*)body.data(), body.size())){
            ostringstream oss;
            oss << "Uncompression failed (size mismatch? header: " << info.size_hdr << ")";
            error(oss.str());
        }
        delete[] buf;

        if(uncompsize != info.size_hdr){
            ostringstream oss;
            oss << "Uncompressed size mismatch "
                << "(header: " << info.size_hdr
                << ", actual: " << uncompsize << ")";
            error(oss.str());
        }
    }

    const char* bool2str(bool b)
    {
        return (b?"true":"false");
    }

    void report(const FCSInfo& info, const string& body)
    {
        if(info.oldver == FCS_OLDVER_NEWFCEU){
            cout << "Emulator: FCEU " << info.newver << endl;
            cout << "Data size (header): " << info.size_hdr << endl;
            cout << "Data size (actual?): " << info.filesize-FCS_HDR_SIZE << endl;
        }
        else if(info.oldver == FCS_OLDVER_FCEUX){
            cout << "Emulator: FCEUX " << info.newver << endl;
            cout << "Compress: " << bool2str(info.comp) << endl;
            if(info.comp){
                cout << "Data size (compress): " << info.compsize << endl;
                cout << "Data size (uncompress): " << info.size_hdr << endl;
            }
            else{
                cout << "Data size: " << info.size_hdr << endl;
            }
        }
        else{
            cout << "Emulator: FCEU (OLD) " << 100*info.oldver << endl;
            cout << "Data size (header): " << info.size_hdr << endl;
            cout << "Data size (actual?): " << info.filesize-FCS_HDR_SIZE << endl;
        }
    }

    void usage()
    {
        error("Usage: fcscheck <FCS>");
    }
}

int main(int argc, char** argv)
{
    if(argc != 2) usage();

    FCSInfo info;
    string body;
    fcs_read(argv[1], info, body);

    chk_comp(info, body);

    report(info, body);

    return 0;
}
