#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include <zlib.h>

#include "binary.hpp"

#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const char* FCS_SIGNATURE = "FCS";
    const uint FCS_HDR_SIZE = 0x10;
    const uchar FCS_OLDVER_FCEUX = 0x58; // 'X'
    const ulong FCS_COMPSIZE_UNCOMP = 0xFFFFFFFF;

    struct FCSInfo{
        ulong newver;
        ulong uncompsize;
        bool comp;
        ulong compsize;
        ulong filesize;
    };

    const char* APP_CMD_COMP = "c";
    const char* APP_CMD_UNCOMP = "u";
    int APP_COMP_LEVEL = Z_BEST_COMPRESSION;

    void warn(const string& msg)
    {
        cerr << msg << endl;
    }

    void error(const string& msg)
    {
        warn(msg);
        exit(1);
    }

    ulong zlib_max_compsize(ulong uncompsize)
    {
        // zlib uncompsize  0.1% + 12Byte vBɔÔ 256Byte 
        return static_cast<ulong>(ceil(uncompsize*1.001)) + 12 + 256;
    }

    void read_chk(istream& in, uchar* buf, streamsize size)
    {
        in.read((char*)buf, size);
        if(!in) error("Incomplete FCS file");
    }

    void read_chk(istream& in, string& buf, streamsize size)
    {
        uchar* tmp = new uchar[size];
        read_chk(in, tmp, size);
        buf.assign((char*)tmp, size);
        delete[] tmp;
    }

    void fcs_read(const string& path, FCSInfo& info, string& body)
    {
        ifstream in(path.c_str(), ios_base::in | ios_base::binary);
        if(!in) error("Can't open input file");
        in.seekg(0, ios_base::end);
        info.filesize = in.tellg();
        in.seekg(0, ios_base::beg);
        if(info.filesize < FCS_HDR_SIZE) error("Incomplete FCS file");

        uchar hdr[FCS_HDR_SIZE];
        read_chk(in, hdr, FCS_HDR_SIZE);
        if(memcmp(hdr+0x00, FCS_SIGNATURE, 3) != 0) error("FCS signature not found");
        uchar oldver = hdr[0x03];
        info.uncompsize = numdec_u32_le(hdr+0x04);
        if(oldver != FCS_OLDVER_FCEUX) error("Not FCS file for FCEUX");

        info.newver = numdec_u32_le(hdr+0x08);
        info.compsize = numdec_u32_le(hdr+0x0C);
        info.comp = (info.compsize != FCS_COMPSIZE_UNCOMP);
        ulong expsize = (info.comp ? info.compsize : info.uncompsize);
        if(expsize+FCS_HDR_SIZE != info.filesize){
            ostringstream oss;
            oss << "Filesize mismatch "
                << "(header: " << expsize+FCS_HDR_SIZE
                << ", actual: " << info.filesize << ")";
            error(oss.str());
        }

        read_chk(in, body, expsize);

        in.close();
    }

    void fcs_write(const string& path, const FCSInfo& info, const uchar* body)
    {
        ofstream out(path.c_str(), ios_base::out | ios_base::binary);
        if(!out) error("Can't open output file");

        uchar hdr[FCS_HDR_SIZE];
        memcpy(hdr+0x00, FCS_SIGNATURE, 3);
        hdr[0x03] = FCS_OLDVER_FCEUX;
        numenc_u32_le(hdr+0x04, info.uncompsize);

        numenc_u32_le(hdr+0x08, info.newver);
        if(info.comp)
            numenc_u32_le(hdr+0x0C, info.compsize);
        else
            numenc_u32_le(hdr+0x0C, FCS_COMPSIZE_UNCOMP);

        ulong bodysize = (info.comp ? info.compsize : info.uncompsize);
        out.write((char*)hdr, FCS_HDR_SIZE);
        out.write((const char*)body, bodysize);

        out.close();
    }

    void fcs_comp(const string& path, const FCSInfo& info, const string& body)
    {
        ulong compsize = zlib_max_compsize(info.uncompsize);
        uchar* body_comp = new uchar[compsize];
        if(Z_OK != compress2(body_comp, &compsize, (const uchar*)body.data(), body.size(), APP_COMP_LEVEL))
            error("Compression failed");

        FCSInfo info_comp = info;
        info_comp.comp = true;
        info_comp.compsize = compsize;
        info_comp.filesize = compsize+FCS_HDR_SIZE;

        fcs_write(path, info_comp, body_comp);

        delete[] body_comp;
    }

    void fcs_uncomp(const string& path, const FCSInfo& info, const string& body)
    {
        ulong uncompsize = info.uncompsize;
        uchar* body_uncomp = new uchar[uncompsize];
        if(Z_OK != uncompress(body_uncomp, &uncompsize, (const uchar*)body.data(), body.size())){
            ostringstream oss;
            oss << "Uncompression failed (filesize mismatch? header: " << info.uncompsize << ")";
            error(oss.str());
        }
        if(uncompsize != info.uncompsize){
            ostringstream oss;
            oss << "Uncompressed size mismatch "
                << "(header: " << info.uncompsize
                << ", actual: " << uncompsize << ")";
            error(oss.str());
        }

        FCSInfo info_uncomp = info;
        info_uncomp.comp = false;
        info_uncomp.compsize = FCS_COMPSIZE_UNCOMP;
        info_uncomp.filesize = uncompsize+FCS_HDR_SIZE;

        fcs_write(path, info_uncomp, body_uncomp);

        delete[] body_uncomp;
    }

    void usage()
    {
        error("Usage: fcscomp <c|u> <input FCS> <output FCS>");
    }
}

int main(int argc, char** argv)
{
    if(argc != 4) usage();
    if(strcmp(argv[1], APP_CMD_COMP)!=0 && strcmp(argv[1], APP_CMD_UNCOMP)!=0) usage();

    FCSInfo info;
    string body;
    fcs_read(argv[2], info, body);

    if(strcmp(argv[1], APP_CMD_COMP) == 0){
        if(info.comp) error("Already compressed");
        fcs_comp(argv[3], info, body);
    }
    else if(strcmp(argv[1], APP_CMD_UNCOMP) == 0){
        if(!info.comp) error("Already uncompressed");
        fcs_uncomp(argv[3], info, body);
    }

    return 0;
}
