#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <cmath>

#include <zlib.h>

#include "binary.hpp"

#include "movie.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const uint FCS_HDR_SIZE = 0x10;
    const ulong FCS_COMPSIZE_UNCOMP = 0xFFFFFFFF;
    struct FCSInfo{
        ulong newver;
        ulong uncompsize;
        bool comp;
        ulong compsize;
        ulong filesize;
    };

    // TASVideosŎgĂFPS
    const double FPS_FCEU[2] = { 60.0, 50.0 };
    const double FPS_FCEUX[2] = { 60.098813897440515532, 50.006977968268290849 };

    const ulong FCEUX_VER_MIN = 20000;

    struct ArgStruct{
        string path_fm2;
        string path_state;
    };

    void warn(const string& msg)
    {
        cerr << msg << endl;
    }

    void error(const string& msg)
    {
        warn(msg);
        exit(1);
    }

    string float2str(double n, int prec)
    {
        ostringstream out;
        out << showpoint << setprecision(prec) << n;
        return out.str();
    }

    bool save_file(const string& path, const string& content)
    {
        ofstream out(path.c_str(), ios_base::out | ios_base::trunc | ios_base::binary);
        if(!out) return false;
        out << content;
        out.close();
        return true;
    }

    void chk_state(FCSInfo& info_fcs, const string& state)
    {
        info_fcs.filesize = state.size();
        if(info_fcs.filesize < FCS_HDR_SIZE) error("State: incomplete FCS");
        const uchar* p = (const uchar*)state.data();

        info_fcs.uncompsize = numdec_u32_le(p+0x04);
        info_fcs.newver = numdec_u32_le(p+0x08);
        info_fcs.compsize = numdec_u32_le(p+0x0C);
        info_fcs.comp = (info_fcs.compsize != FCS_COMPSIZE_UNCOMP);

        if(info_fcs.comp){
            ulong uncompsize = info_fcs.uncompsize;
            uchar* buf = new uchar[uncompsize];
            if(Z_OK != uncompress(buf, &uncompsize, p+FCS_HDR_SIZE, info_fcs.compsize)){
                ostringstream oss;
                oss << "State: uncompression failed (size mismatch? header: " << info_fcs.uncompsize << ")";
                error(oss.str());
            }
            delete[] buf;

            if(uncompsize != info_fcs.uncompsize){
                ostringstream oss;
                oss << "Uncompressed size mismatch "
                    << "(header: " << info_fcs.uncompsize
                    << ", actual: " << uncompsize << ")";
                error(oss.str());
            }
        }
    }

    void frame2time(ulong frame, double fps, string& time)
    {
        ulong csec = (ulong)floor(100.0*frame/fps + 0.5);

        ostringstream out;
        out << setfill('0') << setw(2) << csec/100/(60*60);
        out << ":";
        out << setfill('0') << setw(2) << (csec/100/60)%60;
        out << ":";
        out << setfill('0') << setw(2) << (csec/100)%60;
        out << ".";
        out << setfill('0') << setw(2) << csec%100;
        time = out.str();
    }

    const char* bool2str(bool b)
    {
        return (b?"true":"false");
    }

    const char* from2str(MovieFrom from)
    {
        switch(from){
        //case MOVIE_FROM_RESET: return "Reset"; // FM2Reset-basedeȂ
        case MOVIE_FROM_POWER: return "Power";
        case MOVIE_FROM_STATE: return "State";
        default:
            error("from2str(): invalid argument");
            return NULL;
        }
    }

    void get_active_port(const EmuInputSeq& inputs, bool active[4])
    {
        for(int i = 0; i < 4; ++i) active[i] = false;

        for(EmuInputSeq::const_iterator i = inputs.begin(); i != inputs.end(); ++i){
            const EmuInput& input = *i;
            for(int i = 0; i < 4; ++i)
                if(input.joy[i]) active[i] = true;
        }
    }

    void report(const FM2Info& info, const FCSInfo& info_fcs, const EmuInputSeq& inputs)
    {
        ulong frame = inputs.size();
        double fps_fceu = FPS_FCEU[(info.pal?1:0)];
        double fps_fceux = FPS_FCEUX[(info.pal?1:0)];
        string time_fceu;
        string time_fceux;
        frame2time(frame, fps_fceu, time_fceu);
        frame2time(frame, fps_fceux, time_fceux);
        // FCEU/FCEUX̔(FCEUXłl̏sĂ)
        string emu_name = (info.emu_ver>=FCEUX_VER_MIN ? "FCEUX" : "FCEU");
        string rom_md5_str;
        md5_to_str(info.rom_md5, rom_md5_str);
        bool port[4];
        get_active_port(inputs, port);

        cout << "Emulator: " << emu_name << " " << info.emu_ver << endl;
        cout << "ROM name: " << info.rom_name << endl;
        cout << "ROM MD5: " << rom_md5_str << endl;
        cout << "Recorded from: " << from2str(info.from) << endl;
        cout << "Timing: " << (info.pal?"PAL":"NTSC") << endl;
        cout << "Frames: " << frame << endl;
        cout << "Time (" << float2str(fps_fceu, 5) << " FPS): " << time_fceu << endl;
        cout << "Time (" << float2str(fps_fceux, 5) << " FPS): " << time_fceux << endl;
        cout << "Rerecords: " << info.rerecord << endl;
        cout << "Controller: ";
        for(int i = 0; i < 4; ++i){
            if(i != 0) cout << ", ";
            cout << "P" << 1+i << "=" << bool2str(port[i]);
        }
        cout << endl;

        if(info.from == MOVIE_FROM_STATE){
            cout << "FCS version: " << info_fcs.newver << endl;
            if(info_fcs.comp){
                cout << "FCS data size (compress): " << info_fcs.compsize << endl;
                cout << "FCS data size (uncompress): " << info_fcs.uncompsize << endl;
            }
            else{
                cout << "FCS data size: " << info_fcs.uncompsize << endl;
            }
        }
    }

    void usage()
    {
        error("Usage: fm2check [-os FCS] <FM2>");
    }

    void parse_args(int argc, char** argv, ArgStruct& arg)
    {
        --argc; ++argv;
        while(argc){
            char* str = *argv;
            if(str[0] != '-') break;

            if(strcmp(str+1, "os") == 0){
                --argc; ++argv;
                if(!argc) usage();
                arg.path_state = *argv;
            }
            --argc; ++argv;
        }

        if(argc != 1) usage();
        arg.path_fm2 = *argv;
    }
}

int main(int argc, char** argv)
{
    ArgStruct arg;
    parse_args(argc, argv, arg);

    FM2Info info;
    string state;
    EmuInputSeq inputs;

    ifstream in(arg.path_fm2.c_str(), ios_base::in | ios_base::binary);
    if(!in) error("Can't open input file");
    try{
        fm2_read(in, info, state, inputs);
    }
    catch(const MovieError& e){
        error(e.what());
    }
    in.close();

    if(!arg.path_state.empty()){
        if(state.empty())
            warn("[WARNING] Not state-based movie");
        else if(!save_file(arg.path_state, state))
            warn("[WARNING] Can't write FCS");
    }

    FCSInfo info_fcs;
    if(info.from == MOVIE_FROM_STATE) chk_state(info_fcs, state);

    report(info, info_fcs, inputs);

    return 0;
}
