#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <stdexcept>
#include <cstring>

#include "binary.hpp"

#include "movie.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const char* FCM_SIGNATURE = "FCM\x1A";
    const uint FCM_HDR_SIZE = 0x34;
    const uint FCM_FLAG_FROMRESET = 0x02;
    const uint FCM_FLAG_PAL = 0x04;
    const uint FCM_FLAG_NOSYNCHACK = 0x10;
    enum FCMCmd{
        FCM_CMD_RESET = 0x01,
        FCM_CMD_POWER = 0x02,
        FCM_CMD_FDSINSERT = 0x18,
        FCM_CMD_FDSSELECT = 0x1A,
    };

    const char* FM2_SIGNATURE = "version 3";
    const char FM2_CH_INPUTSEP = '|';
    const char FM2_CH_JOY[8] = { 'A', 'B', 'S', 'T', 'U', 'D', 'L', 'R' };
    const char FM2_CH_NOINPUT[2] = { ' ', '.' };
    enum FM2Cmd{
        FM2_CMD_NONE = 0x00,
        FM2_CMD_RESET = 0x01,
        FM2_CMD_POWER = 0x02,
        FM2_CMD_FDSINSERT = 0x04,
        FM2_CMD_FDSSELECT = 0x08,
    };
    enum FM2Device{
        FM2_DEVICE_GAMEPAD = 0x01,
        FM2_DEVICE_ZAPPER = 0x02,
    };
    enum FM2DeviceEx{
        FM2_DEVICEEX_NONE = 0x00,
    };
    typedef map<string, string> FM2Dict;

    const char* FCS_SIGNATURE = "FCS";
    const uint FCS_HDR_SIZE = 0x10;
    const uchar FCS_OLDVER_FCEUX = 0x58; // 'X'
    const ulong FCS_COMPSIZE_UNCOMP = 0xFFFFFFFF;

    bool hex_to_u8(const string& str, uchar& n)
    {
        if(str.size() < 2) return false;

        uint n2;
        if(!from_string(str.substr(0, 2), n2, hex)) return false;
        n = static_cast<uchar>(n2);
        return true;
    }

    ulong align4(ulong n, uint& n_pad)
    {
        n_pad = (4-(n&0x03))&0x03;
        return n + n_pad;
    }

    void movie_error(const string& msg) { throw MovieError(msg); }

    void read_chk(istream& in, uchar* buf, streamsize size)
    {
        in.read((char*)buf, size);
        if(!in) movie_error("Incomplete file");
    }

    void read_chk(istream& in, string& buf, streamsize size)
    {
        uchar* tmp = new uchar[size];
        read_chk(in, tmp, size);
        buf.assign((char*)tmp, size); // nul܂܂Ă\̂ŒPȂł̓_
        delete[] tmp;
    }

    void read_cstr_chk(istream& in, string& buf)
    {
        stringbuf tmp;
        in.get(tmp, '\0');
        char c;
        in.get(c);
        if(c != '\0')
            movie_error("Incomplete file");
        else
            // in.get(tmp, '\0') 1ǂ߂Ȃ(Ȃ '\0' ɂԂ
            // )ꍇ failbit ̂ŃNABĂ
            // Ȃƌ̃G[`FbNɉeoB
            if(in.fail()) in.clear();
        buf = tmp.str();
    }

    void make_input(EmuInput& input, EmuCmd cmd, const uchar joy[4])
    {
        input.cmd = cmd;
        memcpy(input.joy, joy, 4);
    }

    void get_active_port(const EmuInputSeq& inputs, bool active[4])
    {
        for(int i = 0; i < 4; ++i) active[i] = false;

        for(EmuInputSeq::const_iterator i = inputs.begin(); i != inputs.end(); ++i){
            const EmuInput& input = *i;
            for(int i = 0; i < 4; ++i)
                if(input.joy[i]) active[i] = true;
        }
    }

    EmuCmd fcm_decode_cmd(uint data)
    {
        switch(data){
        case FCM_CMD_RESET: return EMU_CMD_RESET;
        case FCM_CMD_POWER: return EMU_CMD_POWER;
        case FCM_CMD_FDSINSERT: return EMU_CMD_FDSINSERT;
        case FCM_CMD_FDSSELECT: return EMU_CMD_FDSSELECT;
        default: return EMU_CMD_UNKNOWN;
        }
    }

    uint fcm_encode_cmd(EmuCmd cmd)
    {
        switch(cmd){
        case EMU_CMD_RESET: return FCM_CMD_RESET;
        case EMU_CMD_POWER: return FCM_CMD_POWER;
        case EMU_CMD_FDSINSERT: return FCM_CMD_FDSINSERT;
        case EMU_CMD_FDSSELECT: return FCM_CMD_FDSSELECT;
        case EMU_CMD_NONE:
        case EMU_CMD_UNKNOWN:
        default:
            movie_error("fcm_encode_cmd(): invalid argument");
            return 0;
        }
    }

    void fcm_update_joy(uchar joy[4], uint data)
    {
        uint id = (data>>3)&0x03;
        uint button = data&0x07;
        joy[id] ^= 1<<button;
    }

    void fcm_diff_joy(const uchar joy_old[4], const uchar joy_new[4], vector<uint>& data_list)
    {
        for(uint id = 0; id < 4; ++id){
            uchar d = joy_old[id] ^ joy_new[id];
            for(uint button = 0; button < 8; ++button){
                if(d&(1<<button)){
                    uint data = (id<<3) | button;
                    data_list.push_back(data);
                }
            }
        }
    }

    ulong fcm_decode_delta(istream& in, uint size)
    {
        if(!size) return 0;
        uchar buf[3];
        read_chk(in, buf, size);
        return numdec_u_le(buf, size);
    }

    uint fcm_encode_delta(uchar buf[3], ulong delta)
    {
        if(delta == 0){
            return 0;
        }
        else if(0 < delta && delta <= 0xFF){
            buf[0] = delta;
            return 1;
        }
        else if(0xFF < delta && delta <= 0xFFFF){
            numenc_u16_le(buf, delta);
            return 2;
        }
        else if(0xFFFF < delta && delta <= 0xFFFFFF){
            numenc_u24_le(buf, delta);
            return 3;
        }
        else{
            movie_error("Too big delta");
            return 99;
        }
    }

    void fcm_decode_inputs(istream& in, bool from_state, ulong frame, FCMInfo& info, EmuInputSeq& inputs)
    {
        if(from_state){
            info.from = MOVIE_FROM_STATE;
        }
        else{
            // Xe[gx[XłȂꍇAŏByteReset܂PowerR}hłȂ΂ȂȂ
            uchar first;
            in.get((char&)first);
            if(!in) movie_error("Incomplete FCM input data");
            switch(first){
            case 0x81: info.from = MOVIE_FROM_RESET; break;
            case 0x82: info.from = MOVIE_FROM_POWER; break;
            default: movie_error("No Power/Reset command at beginning of non-state-based movie");
            }
        }

        ulong f = 0;
        EmuCmd cmd = EMU_CMD_NONE;
        uchar joy[4] = { 0, 0, 0, 0 };
        for(;;){
            uchar update;
            in.get((char&)update);
            if(!in) movie_error("Incomplete FCM input data");

            bool is_cmd = update&0x80;
            uint deltasize = (update>>5)&0x03;
            uint data = update&0x1F;

            // delta t[̓͂𐶐
            ulong delta = fcm_decode_delta(in, deltasize);
            if(delta){
                // R}h͍ŏ̃t[݂̂ɓKp
                EmuInput input;
                make_input(input, cmd, joy);
                inputs.push_back(input);
                cmd = input.cmd = EMU_CMD_NONE;
                // c̃t[𐶐
                for(ulong i = 0; i < delta-1; ++i)
                    inputs.push_back(input);
                f += delta;
            }

            // ͂̍XV
            if(is_cmd){
                // ܂EOF}[Jǂ`FbN
                if(data == 0) break;

                // 1Fɕ̃R}h܂߂Ă͂ȂȂ(FM2ł͂͋ĂȂ̂)
                if(cmd != EMU_CMD_NONE){
                    ostringstream oss;
                    oss << "Frame " << f << ": multiple commands in 1 frame";
                    movie_error(oss.str());
                }

                cmd = fcm_decode_cmd(data);
                if(cmd == EMU_CMD_UNKNOWN){
                    ostringstream oss;
                    oss << "Frame " << f << ": unknown command";
                    movie_error(oss.str());
                }
            }
            else{
                fcm_update_joy(joy, data);
            }
        }

        if(frame != inputs.size()){
            ostringstream oss;
            oss << "Frame count mismatch (header: " << frame << ", actual: " << inputs.size() << ")";
            movie_error(oss.str());
        }
    }

    void fcm_encode_inputs(ostream& out, const FCMInfo& info, const EmuInputSeq& inputs)
    {
        switch(info.from){
        case MOVIE_FROM_RESET: out.put(0x81); break;
        case MOVIE_FROM_POWER: out.put(0x82); break;
        case MOVIE_FROM_STATE: break;
        default: movie_error("invalid starting point");
        }

        ulong delta = 0;
        uchar delta_buf[3];
        uchar joy[4] = { 0, 0, 0, 0 };
        for(ulong f = 0; f < inputs.size(); ++f){
            const EmuInput& input = inputs[f];

            if(input.cmd != EMU_CMD_NONE){
                uint data = fcm_encode_cmd(input.cmd);
                uint deltasize = fcm_encode_delta(delta_buf, delta);
                uchar update = 0x80 | (deltasize<<5) | data;
                out.put(update);
                if(deltasize) out.write((char*)delta_buf, deltasize);
                delta = 0;
            }

            if(memcmp(input.joy, joy, 4) != 0){
                // ͂̍Ƃ
                vector<uint> data_list;
                fcm_diff_joy(joy, input.joy, data_list);
                // ŏ update ̂ delta t
                vector<uint>::const_iterator i = data_list.begin();
                uint deltasize = fcm_encode_delta(delta_buf, delta);
                uchar update = (deltasize<<5) | *i;
                out.put(update);
                if(deltasize) out.write((char*)delta_buf, deltasize);
                delta = 0;
                // c update o
                for(++i; i != data_list.end(); ++i)
                    out.put(*i);
                // ͏ԂXV
                memcpy(joy, input.joy, 4);
            }

            ++delta;
        }

        // EOF}[Jo
        uint deltasize = fcm_encode_delta(delta_buf, delta);
        uchar update = 0x80 | (deltasize<<5);
        out.put(update);
        if(deltasize) out.write((char*)delta_buf, deltasize);
    }

    EmuCmd fm2_decode_cmd(uint cmd_fm2)
    {
        switch(cmd_fm2){
        case FM2_CMD_NONE: return EMU_CMD_NONE;
        case FM2_CMD_RESET: return EMU_CMD_RESET;
        case FM2_CMD_POWER: return EMU_CMD_POWER;
        case FM2_CMD_FDSINSERT: return EMU_CMD_FDSINSERT;
        case FM2_CMD_FDSSELECT: return EMU_CMD_FDSSELECT;
        default: return EMU_CMD_UNKNOWN;
        }
    }

    uint fm2_encode_cmd(EmuCmd cmd)
    {
        switch(cmd){
            case EMU_CMD_NONE: return FM2_CMD_NONE;
            case EMU_CMD_RESET: return FM2_CMD_RESET;
            case EMU_CMD_POWER: return FM2_CMD_POWER;
            case EMU_CMD_FDSINSERT: return FM2_CMD_FDSINSERT;
            case EMU_CMD_FDSSELECT: return FM2_CMD_FDSSELECT;
            case EMU_CMD_UNKNOWN:
            default:
                movie_error("fm2_encode_cmd(): invalid argument");
                return 0xFF;
        }
    }

    // "RLDUTSBA"
    uchar fm2_decode_joy(const string& str)
    {
        if(str.size() != 8) movie_error("Invalid joy string");

        uchar joy = 0;
        for(int i = 0; i < 8; ++i){
            char c = str[i];
            for(ulong j = 0; j < NELEMS(FM2_CH_NOINPUT); ++j)
                if(c == FM2_CH_NOINPUT[j]) goto NEXT;
            // FCEUXł͖͂\ȊO͑Sē͂ƂĈAꉞ`FbN
            if(c == FM2_CH_JOY[7-i])
                joy |= (1<<(7-i));
            else
                movie_error("Invalid joy char");
NEXT:
            ;
        }

        return joy;
    }

    string fm2_encode_joy(uchar joy)
    {
        string result;
        for(int i = 0; i < 8; ++i)
            result += (joy&(1<<(7-i))) ? FM2_CH_JOY[7-i] : '.';
        return result;
    }

    ulong fm2_decode_u32(const string& str)
    {
        ulong n;
        if(!from_string(str, n)) movie_error("Can't parse number");
        return n;
    }

    bool fm2_decode_bool(const string& str)
    {
        return fm2_decode_u32(str) != 0;
    }

    string fm2_decode_blob(const string& str)
    {
        if(str.size() < 2) movie_error("Can't parse blob");

        string result;
        if(str[0] == '0' && (str[1] == 'x' || str[1] == 'X')){ // hex
            if(str.size()&1) movie_error("Can't parse hex blob");

            for(ulong i = 1; i < str.size()/2; ++i){
                uchar n;
                if(!hex_to_u8(str.substr(2*i, 2), n)) movie_error("Can't parse hex blob");
                result += static_cast<char>(n);
            }
        }
        else{ // base64
            if(str.size() < 7) movie_error("Can't parse blob");
            if(str.substr(0, 7) != "base64:") movie_error("Can't parse blob");
            if(str.size() == 7) return "";

            if(!base64_decode(str.substr(7), result))
                movie_error("Can't decode base64 blob");
        }

        return result;
    }

    string fm2_encode_blob(const string& str)
    {
        string result("base64:");
        string b64;
        base64_encode(str, b64);
        result += b64;
        return result;
    }

    void fm2_parse_input(const string& line, EmuInputSeq& inputs)
    {
        vector<string> v;
        split(v, line, FM2_CH_INPUTSEP);
        bool fourscore = false;
        if(v.size() == 6)
            fourscore = false;
        else if(v.size() == 8)
            fourscore = true;
        else
            movie_error("Can't parse line");

        EmuCmd cmd;
        uchar joy[4] = { 0, 0, 0, 0 };
        uint cmd_fm2;
        if(!from_string(v[1], cmd_fm2)) movie_error("Invalid command");
        if((cmd = fm2_decode_cmd(cmd_fm2)) == EMU_CMD_UNKNOWN) movie_error("Unknown command");
        for(int i = 0; i < (fourscore?4:2); ++i)
            joy[i] = fm2_decode_joy(v[2+i]);

        EmuInput input;
        make_input(input, cmd, joy);
        inputs.push_back(input);
    }

    string fm2_encode_input(const EmuInput& input, bool fourscore)
    {
        ostringstream out;
        out << "|" << fm2_encode_cmd(input.cmd) << "|";
        for(int i = 0; i < (fourscore?4:2); ++i)
            out << fm2_encode_joy(input.joy[i]) << "|";
        out << "|";
        return out.str();
    }

    void fm2_parse_pair(const string& line, FM2Dict& dict)
    {
        vector<string> v;
        split(v, line, 1);
        if(v.size() == 1)
            dict[v[0]] = "";
        else if(v.size() == 2)
            dict[v[0]] = v[1];
        else
            movie_error("fm2_parse_pair(): Internal error");
    }
}

MovieError::MovieError(const string& msg) : runtime_error(msg) {}

void fcm_read(istream& in, FCMInfo& info, string& state, EmuInputSeq& inputs)
{
    in.seekg(0, ios_base::end);
    ulong filesize = in.tellg();
    in.seekg(0, ios_base::beg);
    if(filesize <= FCM_HDR_SIZE) movie_error("Incomplete FCM file");

    uchar hdr[FCM_HDR_SIZE];
    read_chk(in, hdr, FCM_HDR_SIZE);
    if(memcmp(hdr+0x00, FCM_SIGNATURE, 4) != 0) movie_error("FCM signature not found");
    ulong fcm_ver = numdec_u32_le(hdr+0x04);
    if(fcm_ver != 2) movie_error("Not FCMv2 file");

    uchar flag = hdr[0x08];
    bool from_state = !(flag&FCM_FLAG_FROMRESET);
    info.pal = flag&FCM_FLAG_PAL;
    info.synchack = !(flag&FCM_FLAG_NOSYNCHACK);

    ulong frame = numdec_u32_le(hdr+0x0C);
    info.rerecord = numdec_u32_le(hdr+0x10);

    ulong bodysize = numdec_u32_le(hdr+0x14);
    ulong stateoff = numdec_u32_le(hdr+0x18);
    ulong bodyoff = numdec_u32_le(hdr+0x1C);
    if(stateoff >= bodyoff) movie_error("Invalid FCM file (stateoff >= bodyoff)");
    if(bodyoff + bodysize > filesize) movie_error("Incomplete FCM file");
    // bodyoff + bodysize < filesize ƂȂ(ɃS~tĂ)ꍇ邪AFCEU̎dlȂ̂ŋe
    ulong statesize = bodyoff - stateoff;
    if(statesize < FCS_HDR_SIZE) movie_error("State: incomplete FCS");

    md5_from_raw(info.rom_md5, hdr+0x20);
    info.emu_ver = numdec_u32_le(hdr+0x30);

    read_cstr_chk(in, info.rom_name);
    read_cstr_chk(in, info.metadata);

    state.clear();
    inputs.clear();

    /**
     * FCEU͋ZNVFCS̃f[^TCY𐳂o͂Ȃ̂ŁAFCS
     * wb_͂ĂɂȂȂBŁAstateoff, bodyoff FCSTCY
     * B̕@FCSƓ̓f[^̊ԂɃMbvꍇFCS
     * ɃS~tĂ܂B
     */
    in.seekg(stateoff, ios_base::beg);
    read_chk(in, state, statesize);
    if(memcmp(state.data()+0x00, FCS_SIGNATURE, 3) != 0) movie_error("State: FCS signature not found");

    /**
     * FCEU̎dlA̓f[^ɃS~tĂ邱Ƃ̂ŁAꉞS
     * ~݂̂𔲂oăfR[h
     */
    string body;
    in.seekg(bodyoff, ios_base::beg);
    read_chk(in, body, bodysize);
    istringstream body_in(body);
    fcm_decode_inputs(body_in, from_state, frame, info, inputs);
}

void fcm_write(ostream& out, const FCMInfo& info, const string& state, const EmuInputSeq& inputs)
{
    // Xe[gx[Xł邩ۂ킸Xe[g͕Kv
    if(state.size() < FCS_HDR_SIZE) movie_error("State: incomplete FCS");
    if(memcmp(state.data()+0x00, FCS_SIGNATURE, 3) != 0) movie_error("State: FCS signature not found");

    ostringstream body_out;
    fcm_encode_inputs(body_out, info, inputs);
    string body = body_out.str();

    // ꉞ string nulĂ\l
    size_t rom_name_len = strlen(info.rom_name.c_str());
    size_t metadata_len = strlen(info.metadata.c_str());
    // Xe[gƓ̓f[^̃ItZbg4ByteEɍ킹(XL}0pfBO)
    // FCEU FCEUI_SaveMovie() łl̏sĂ
    uint n_pad[2];
    ulong stateoff = align4(FCM_HDR_SIZE + (rom_name_len+1) + (metadata_len+1), n_pad[0]);
    ulong bodyoff = align4(stateoff + state.size(), n_pad[1]);

    uchar hdr[FCM_HDR_SIZE];
    memcpy(hdr+0x00, FCM_SIGNATURE, 4);
    numenc_u32_le(hdr+0x04, 2);

    uchar flag = 0;
    if(info.from != MOVIE_FROM_STATE) flag |= FCM_FLAG_FROMRESET;
    if(info.pal) flag |= FCM_FLAG_PAL;
    if(!info.synchack) flag |= FCM_FLAG_NOSYNCHACK;
    hdr[0x08] = flag;
    memset(hdr+0x09, 0, 3);

    numenc_u32_le(hdr+0x0C, inputs.size());
    numenc_u32_le(hdr+0x10, info.rerecord);

    numenc_u32_le(hdr+0x14, body.size());
    numenc_u32_le(hdr+0x18, stateoff);
    numenc_u32_le(hdr+0x1C, bodyoff);

    md5_to_raw(info.rom_md5, hdr+0x20);
    numenc_u32_le(hdr+0x30, info.emu_ver);

    uchar pad[3] = { 0, 0, 0 };
    out.write((char*)hdr, FCM_HDR_SIZE);
    out.write(info.rom_name.c_str(), rom_name_len+1);
    out.write(info.metadata.c_str(), metadata_len+1);
    if(n_pad[0]) out.write((char*)pad, n_pad[0]);
    out.write(state.data(), state.size());
    if(n_pad[1]) out.write((char*)pad, n_pad[1]);
    out.write(body.data(), body.size());
}

/**
 * port2 ͍̓͂̂Ƃ떳
 *
 * comment, subtitle ɂĂ͍̂Ƃ떳
 * (L[ɑ΂ĕ̒l݂P[X̂߂ǂ̂)
 */
void fm2_read(istream& in, FM2Info& info, string& state, EmuInputSeq& inputs)
{
    string line;
    ulong linenum = 1;
    if(!getline_u(in, line)) movie_error("Incomplete FM2 file");
    if(line != FM2_SIGNATURE) movie_error("FM2 signature not found");

    state.clear();
    inputs.clear();

    FM2Dict dict;
    while(getline_u(in, line)){
        ++linenum;
        // FM2ɂ󔒕̓Xy[Xƃ^û݂A܂ŌɈKvȂƎv
        strip(line);
        if(line.empty()) continue;

        try{
            if(line[0] == FM2_CH_INPUTSEP)
                fm2_parse_input(line, inputs);
            else
                fm2_parse_pair(line, dict);
        }
        catch(const MovieError& e){
            ostringstream oss;
            oss << "Line " << linenum << ": " << e.what();
            movie_error(oss.str());
        }
    }

    bool active[4];
    get_active_port(inputs, active);

    FM2Dict::const_iterator i;
    if((i = dict.find("emuVersion")) == dict.end()) movie_error("emuVersion not found");
    info.emu_ver = fm2_decode_u32((*i).second);
    if((i = dict.find("romFilename")) == dict.end()) movie_error("romFilename not found");
    info.rom_name = (*i).second;
    if((i = dict.find("romChecksum")) == dict.end()) movie_error("romChecksum not found");
    string md5_raw = fm2_decode_blob((*i).second);
    md5_from_raw(info.rom_md5, (const uchar*)md5_raw.data());
    if((i = dict.find("guid")) == dict.end()) movie_error("guid not found");
    if(!uuid_from_str(info.uuid, (*i).second))
        movie_error("Can't parse UUID");

    bool fourscore = false;
    if((i = dict.find("fourscore")) != dict.end())
        fourscore = fm2_decode_bool((*i).second);
    if(!fourscore){
        if(active[2] || active[3]) movie_error("fourscore must be true");
        ulong port0, port1;
        if((i = dict.find("port0")) == dict.end()) movie_error("port0 not found");
        port0 = fm2_decode_u32((*i).second);
        if((i = dict.find("port1")) == dict.end()) movie_error("port1 not found");
        port1 = fm2_decode_u32((*i).second);
        if(port0 != FM2_DEVICE_GAMEPAD || port1 != FM2_DEVICE_GAMEPAD)
            movie_error("Only gamepad is supported. Sorry.");
    }
    if((i = dict.find("port2")) == dict.end()) movie_error("port2 not found");
    ulong port2 = fm2_decode_u32((*i).second);
    if(port2 != FM2_DEVICEEX_NONE)
        movie_error("FCExp device is not supported. Sorry.");

    info.pal = false;
    if((i = dict.find("palFlag")) != dict.end())
        info.pal = fm2_decode_bool((*i).second);
    info.rerecord = 0;
    if((i = dict.find("rerecordCount")) != dict.end())
        info.rerecord = fm2_decode_u32((*i).second);
    if((i = dict.find("savestate")) != dict.end()){
        state = fm2_decode_blob((*i).second);
        if(state.size() < FCS_HDR_SIZE) movie_error("State: incomplete FCS");
        const uchar* p = (const uchar*)state.data();
        if(memcmp(p+0x00, FCS_SIGNATURE, 3) != 0) movie_error("State: FCS signature not found");
        if(p[0x03] != FCS_OLDVER_FCEUX) movie_error("State: not FCS for FCEUX");
        /**
         * FCEUX͏ɐTCYFCSwb_ɏނ̂ŁATCY`FbN
         */
        ulong uncompsize = numdec_u32_le(p+0x04);
        ulong compsize = numdec_u32_le(p+0x0C);
        ulong expsize = (compsize==FCS_COMPSIZE_UNCOMP ? uncompsize : compsize);
        if(expsize+FCS_HDR_SIZE != state.size()){
            ostringstream oss;
            oss << "FCS size mismatch (header: " << expsize+FCS_HDR_SIZE << ", actual: " << state.size() << ")";
            movie_error(oss.str());
        }
    }

    info.from = (state.empty() ? MOVIE_FROM_POWER : MOVIE_FROM_STATE);
}

void fm2_write(ostream& out, const FM2Info& info, const string& state, const EmuInputSeq& inputs)
{
    static const char CRLF[] = "\x0D\x0A";

    if(info.from == MOVIE_FROM_STATE){
        if(state.size() < FCS_HDR_SIZE) movie_error("State: incomplete FCS");
        const uchar* p = (const uchar*)state.data();
        if(memcmp(p+0x00, FCS_SIGNATURE, 3) != 0) movie_error("State: FCS signature not found");
        if(p[0x03] != FCS_OLDVER_FCEUX) movie_error("State: not FCS for FCEUX");
    }

    bool active[4];
    get_active_port(inputs, active);
    bool fourscore = (active[2] || active[3]);
    uchar md5_raw[16];
    md5_to_raw(info.rom_md5, md5_raw);
    string md5_blob = fm2_encode_blob(string((char*)md5_raw, 16));
    string uuid_str;
    uuid_to_str(info.uuid, uuid_str);

    out << FM2_SIGNATURE << CRLF;
    out << "emuVersion " << info.emu_ver << CRLF;
    out << "rerecordCount " << info.rerecord << CRLF;
    out << "palFlag " << (info.pal?1:0) << CRLF;
    out << "romFilename " << info.rom_name << CRLF;
    out << "romChecksum " << md5_blob << CRLF;
    out << "guid " << uuid_str << CRLF;
    out << "fourscore " << (fourscore?1:0) << CRLF;
    out << "port0 " << FM2_DEVICE_GAMEPAD << CRLF;
    out << "port1 " << FM2_DEVICE_GAMEPAD << CRLF;
    out << "port2 " << FM2_DEVICEEX_NONE << CRLF;
    if(info.from == MOVIE_FROM_STATE)
        out << "savestate " << fm2_encode_blob(state) << CRLF;

    for(EmuInputSeq::const_iterator i = inputs.begin(); i != inputs.end(); ++i)
        out << fm2_encode_input(*i, fourscore) << CRLF;
}
