#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <vector>
#include <stdexcept>
#include <cstring>

#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    const char* CH_WHITESPACE = " \t\x0A\x0D\x0C";

    const char BASE64_CH[64] = {
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
        'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
        'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        '+', '/',
    };
    const char BASE64_CH_PAD = '=';
    const int BASE64_VALUE[0x100] = {
        /* 0x00 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0x10 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0x20 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
        /* 0x30 */ 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
        /* 0x40 */ -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
        /* 0x50 */ 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
        /* 0x60 */ -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
        /* 0x70 */ 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
        /* 0x80 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0x90 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0xA0 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0xB0 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0xC0 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0xD0 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0xE0 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
        /* 0xF0 */ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    };

    void split(vector<string>& v, const string& str, int* maxsplit)
    {
        if(maxsplit && *maxsplit < 0) throw invalid_argument("split(): negative maxsplit");

        v.clear();
        string s = str;
        strip(s);

        int n = 0;
        string::size_type pos_orig = 0;
        string::size_type pos;
        while((!maxsplit || n < *maxsplit) &&
              (pos = s.find_first_of(CH_WHITESPACE, pos_orig)) != string::npos){
            if(pos_orig != pos){
                v.push_back(s.substr(pos_orig, pos-pos_orig));
                ++n;
            }
            pos_orig = pos+1;
            if(pos_orig > s.size()-1) break;
        }

        if(pos_orig <= s.size()-1)
            v.push_back(s.substr(pos_orig));
    }

    void split(vector<string>& v, const string& str, const string& sep, int* maxsplit)
    {
        if(maxsplit && *maxsplit < 0) throw invalid_argument("split(): negative maxsplit");

        v.clear();

        int n = 0;
        string::size_type sep_size = sep.size();
        string::size_type pos_orig = 0;
        string::size_type pos;
        while((!maxsplit || n < *maxsplit) &&
              (pos = str.find(sep, pos_orig)) != string::npos){
            v.push_back(pos_orig==pos ? string("") : str.substr(pos_orig,pos-pos_orig));
            ++n;
            pos_orig = pos+sep_size;
            if(pos_orig > str.size()-1) break;
        }

        v.push_back(pos_orig==str.size() ? string("") : str.substr(pos_orig));
    }

    string u8_to_hex(uchar n, bool upper)
    {
        ostringstream out;
        out << hex;
        if(upper) out << uppercase;
        out << setfill('0') << setw(2) << static_cast<uint>(n);
        return out.str();
    }

    bool hex_to_u8(const string& str, uchar& n)
    {
        if(str.size() < 2) return false;

        uint n2;
        if(!from_string(str.substr(0, 2), n2, hex)) return false;
        n = static_cast<uchar>(n2);
        return true;
    }

    void base64_enc_unit(const uchar plain[3], char enc[4])
    {
        enc[0] = BASE64_CH[(plain[0]>>2)&0x3F];
        enc[1] = BASE64_CH[((plain[0]<<4)|(plain[1]>>4))&0x3F];
        enc[2] = BASE64_CH[((plain[1]<<2)|(plain[2]>>6))&0x3F];
        enc[3] = BASE64_CH[plain[2]&0x3F];
    }

    void base64_dec_unit(const char enc[4], uchar plain[3])
    {
        uchar val[4];
        for(int i = 0; i < 4; ++i)
            val[i] = BASE64_VALUE[static_cast<int>(enc[i])];
        plain[0] = (val[0]<<2) | ((val[1]>>4)&0x03);
        plain[1] = (val[1]<<4) | ((val[2]>>2)&0x0F);
        plain[2] = (val[2]<<6) | val[3];
    }
}

void lstrip(string& str)
{
    lstrip(str, CH_WHITESPACE);
}

void lstrip(string& str, const char* ch)
{
    string::size_type pos = str.find_first_not_of(ch);
    if(pos != string::npos && pos > 0)
        str.erase(0, pos);
}

void rstrip(string& str)
{
    rstrip(str, CH_WHITESPACE);
}

void rstrip(string& str, const char* ch)
{
    string::size_type pos = str.find_last_not_of(ch);
    if(pos != string::npos && pos < str.size()-1)
        str.erase(pos+1, string::npos);
}

void strip(string& str)
{
    strip(str, CH_WHITESPACE);
}

void strip(string& str, const char* ch)
{
    lstrip(str, ch);
    rstrip(str, ch);
}

void split(vector<string>& v, const string& str)
{
    split(v, str, static_cast<int*>(NULL));
}

void split(vector<string>& v, const string& str, int maxsplit)
{
    split(v, str, &maxsplit);
}

void split(vector<string>& v, const string& str, char sep)
{
    split(v, str, string("")+sep);
}

void split(vector<string>& v, const string& str, char sep, int maxsplit)
{
    split(v, str, string("")+sep, maxsplit);
}

void split(vector<string>& v, const string& str, const string& sep)
{
    split(v, str, sep, static_cast<int*>(NULL));
}

void split(vector<string>& v, const string& str, const string& sep, int maxsplit)
{
    split(v, str, sep, &maxsplit);
}

string join(const string& sep, const vector<string>& v)
{
    string str("");
    for(vector<string>::const_iterator i = v.begin(); i != v.end(); ++i){
        if(i != v.begin()) str += sep;
        str += *i;
    }
    return str;
}

/**
 * ԃtÖɂ std::getline() Ƌ킹KvB
 *
 * std::getline() s𔭌łȂ܂EOFɒBꍇA
 *   * string 1łǂݍłAeofbit ͗ failbit ͗Ȃ
 *   * string 1ǂݍłȂ΁Aeofbit, failbit 
 */
istream& getline_u(istream& in, string& line)
{
    static const char LF = '\x0A';
    static const char CR = '\x0D';

    line.clear();

    for(;;){
        char c;
        in.get(c);

        if(in.eof()){
            if(line.empty())
                in.setstate(ios_base::eofbit | ios_base::failbit);
            else
                in.clear(ios_base::eofbit);
            return in;
        }

        switch(c){
        case LF:
            return in;
        case CR:
            /**
             * peek() EOFoꍇAeofbit Ă܂(failbit 
             * Ȃ)悤Ȃ̂ŁAstd::getline() Ƌ킹邽
             * clear()  eofbit 낵ĂB
             */
            int c_next = in.peek();
            if(c_next == LF)
                in.get();
            else if(in.eof()) in.clear();
            return in;
        default:
            line += c;
        }
    }
}

void md5_from_raw(MD5& md5, const uchar* raw)
{
    memcpy(md5.b, raw, 16);
}

void md5_to_raw(const MD5& md5, uchar* raw)
{
    memcpy(raw, md5.b, 16);
}

void md5_to_str(const MD5& md5, string& str)
{
    str.clear();
    for(int i = 0; i < 16; ++i) str += u8_to_hex(md5.b[i], false);
}

void uuid_from_raw(UUID& uuid, const uchar* raw)
{
    memcpy(uuid.b, raw, 16);
}

void uuid_to_raw(const UUID& uuid, uchar* raw)
{
    memcpy(raw, uuid.b, 16);
}

bool uuid_from_str(UUID& uuid, const string& str)
{
    if(str.size() != 16*2 + 4) return false;
    if(!(str[8] == '-' && str[13] == '-' && str[18] == '-' && str[23] == '-')) return false;

    vector<string> v;
    split(v, str, '-');
    if(v.size() != 5) return false;
    string s = join("", v);

    for(int i = 0; i < 16; ++i)
        if(!hex_to_u8(s.substr(2*i, 2), uuid.b[i])) return false;

    return true;
}

void uuid_to_str(const UUID& uuid, string& str)
{
    str.clear();
    for(int i = 0; i < 4; ++i) str += u8_to_hex(uuid.b[i], true);
    str += '-';
    for(int i = 4; i < 6; ++i) str += u8_to_hex(uuid.b[i], true);
    str += '-';
    for(int i = 6; i < 8; ++i) str += u8_to_hex(uuid.b[i], true);
    str += '-';
    for(int i = 8; i < 10; ++i) str += u8_to_hex(uuid.b[i], true);
    str += '-';
    for(int i = 10; i < 16; ++i) str += u8_to_hex(uuid.b[i], true);
}

void base64_encode(const string& src, string& dst)
{
    string::size_type size = src.size();
    dst.clear();

    int n = 0;
    const char* p = src.data();
    char plain[3];
    char enc[4];
    while(size--){
        plain[n++] = *p++;
        if(n == 3){
            base64_enc_unit((uchar*)plain, enc);
            dst.append(enc, 4);
            n = 0;
        }
    }

    if(n){
        for(int i = n; i < 3; ++i)
            plain[i] = 0;
        base64_enc_unit((uchar*)plain, enc);
        for(int i = n+1; i < 4; ++i)
            enc[i] = BASE64_CH_PAD;
        dst.append(enc, 4);
    }
}

bool base64_decode(const string& src, string& dst)
{
    string::size_type size = src.size();
    if(size%4) return false;
    dst.clear();

    int n = 0;
    const char* p = src.data();
    char plain[3];
    char enc[4];
    for( ; size > 0; --size){
        if(BASE64_VALUE[static_cast<int>(*p)] == -1) break;
        enc[n++] = *p++;
        if(n == 4){
            base64_dec_unit(enc, (uchar*)plain);
            dst.append(plain, 3);
            n = 0;
        }
    }

    if(size > 2){
        return false;
    }
    else if(size > 0){
        for(ulong i = 4-size; i < 4; ++i){
            if(*p++ != BASE64_CH_PAD) return false;
            enc[i] = BASE64_CH[0];
        }
        base64_dec_unit(enc, (uchar*)plain);
        dst.append(plain, (size==2 ? 1 : 2));
    }

    return true;
}
